"This file defines a linked-list ADT"

"the ancestor of all lists"
globals _AddSlotsIfAbsent: ( |
  listTraits = ( ).
| ).
listTraits _DefineSlots: ( |
  parent* = collectionTraits.

  "add a new cons cell on to the front of the receiver list"
  addToFront: newFirst = (
    (consProto clone first: newFirst) rest: self ).

  "evaluate aBlock on each element in the list"
  do: aBlock = (
    | link |
    link: self.
    [link isEmpty] whileFalse: [ 
      aBlock value: link first. 
      link: link rest. ].
    nil ).

  at: index Put: newValue IfOutOfBounds: outOfBoundsBlock = (
    | link. pos <- 0. |
    index >= 0 ifTrue: [
      link: self.
      [link isEmpty] whileFalse: [ 
        pos = index ifTrue: [
	  "we found the right link to update"
	  link first: newValue.
	  ^ self
	].
	"keep scanning"
        link: link rest.
	pos: pos + 1.
      ].
    ].
    "the index was out of bounds"
    outOfBoundsBlock value ).

  "printing behavior for lists"
  printString = (
    | str <- 'list{'. first <- true. |
    do: [|:elem| 
      first ifTrue: [ first: false ] False: [ str: str + ', '. ].
      str: str + elem printString. ].
    str + '}' ).
| ).

"the parent of cons cells"
globals _AddSlotsIfAbsent: ( |
  consTraits = ( ).
| ).
consTraits _DefineSlots: ( |
  parent* = listTraits.

  isEmpty = false.
| ).

"the prototypical cons cell, 'hidden' inside the listTraits object so
 that it's less visible to the rest of the world"
listTraits _AddSlots: ( |
  consProto = ( |
    parent* = consTraits.
    first.
    rest.
  | ).
| ).

"update nil with behavior to act like the empty list"
nil _AddSlots: ( |
  parent* = listTraits.

  "put empty-list operations here"

  isEmpty = true.
| ).

"add public operations to all objects"
objectTraits _AddSlots: ( |
  "create a new list which is the receiver element followed by the
   argument list"
  cons: rest = ( rest addToFront: self ).
| ).


"singly linked lists are bad at adding to the end.  so to construct a
list in left-to-right order, define this helper object that allows
constant-time adding to the end of the list.  see map: in
collection.self for a simple use"

globals _AddSlotsIfAbsent: ( |
  listBuilderTraits = ( ).
| ).
listBuilderTraits _DefineSlots: ( |
  parent* = objectTraits.

  "allow adding to the front and back of the list being built"
  addToFront: elem = ( 
    list: list addToFront: elem.
    last isEmpty ifTrue: [ last: list. ]. ).

  addToBack: elem = (
    last isEmpty 
      ifTrue: [ addToFront: elem ]
       False: [ | newLast |
         newLast: elem cons: nil.
	 last rest: newLast.
	 last: newLast. ] ).
| ).

globals _AddSlotsIfAbsent: ( |
  listBuilder = ( ).
| ).
listBuilder _DefineSlots: ( |
  parent* = listBuilderTraits.

  "send 'list' to get the list object that's been built"
  list.

  "key idea: builder keeps track of the last cons cell in the list"
  last.
| ).
